# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import base64
import ensure
import json
import logging
import os
import socket
import typing

from Crypto.Cipher import AES
import Crypto.Util.Padding
import requests

import ci.log
from ci.util import urljoin
import model
import model.concourse
import model.secret
import model.secrets_server


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


def get_secret_cfg_from_env_if_available(
    key_env_var='SECRET_KEY',
    cipher_algorithm_env_var='SECRET_CIPHER_ALGORITHM',
) -> typing.Optional[model.secret.SecretData]:
    if not all((name in os.environ for name in (key_env_var, cipher_algorithm_env_var))):
        return None

    secret_key = base64.b64decode(os.environ.get(key_env_var).encode('utf-8'))
    cipher_algorithm = model.secret.Cipher(os.environ.get(cipher_algorithm_env_var))
    secret = model.secret.SecretData(key=secret_key, cipher_algorithm=cipher_algorithm)

    return secret


class SecretsServerClient:
    @staticmethod
    def from_env(
        endpoint_env_var='SECRETS_SERVER_ENDPOINT',
        concourse_secret_env_var='SECRETS_SERVER_CONCOURSE_CFG_NAME',
        cache_file='SECRETS_SERVER_CACHE'
    ):
        if cache_file not in os.environ:
            if not all(map(
                        lambda e: e in os.environ,
                        (endpoint_env_var, concourse_secret_env_var)
                        )):
                raise ValueError('the following environment variables must be defined: {v}'.format(
                    v=', '.join((endpoint_env_var, concourse_secret_env_var))
                ))
        cache_file = os.environ.get(cache_file, None)

        secret = get_secret_cfg_from_env_if_available()

        return SecretsServerClient(
            endpoint_url=os.environ.get(endpoint_env_var),
            concourse_secret_name=os.environ.get(concourse_secret_env_var),
            cache_file=cache_file,
            secret=secret,
        )

    @staticmethod
    def default():
        # hardcode default endpoint name (usually injected via env (see above))
        default_secrets_server_hostname = 'secrets-server.concourse.svc.cluster.local'
        try:
            socket.getaddrinfo(default_secrets_server_hostname, 80)
        except socket.gaierror:
            raise ValueError('secrets-server not accessible (are you running in ci-cluster?)')
        # also hardcode default url path (usually injected via env)

        # secret will be read from environemt variables
        secret = get_secret_cfg_from_env_if_available()
        if current_team := os.environ.get('CONCOURSE_CURRENT_TEAM'):
            default_secrets_path = model.secrets_server._org_based_secret_url_path(
                target_secret_name=model.concourse.cfg_name_from_team(current_team),
                secret_cfg_name=model.concourse.secret_cfg_name_for_team(current_team),
            )
        else:
            logger.warning(
                'CONCOURSE_CURRENT_TEAM not found in environment. Needed for secret retrieval.',
            )

        return SecretsServerClient(
            endpoint_url=f'http://{default_secrets_server_hostname}',
            concourse_secret_name=default_secrets_path,
            cache_file=None,
            secret=secret,
        )

    def __init__(
        self,
        endpoint_url,
        concourse_secret_name,
        cache_file=None,
        secret: model.secret.SecretData = None,
    ):
        self.url = endpoint_url
        self.concourse_secret_name = concourse_secret_name
        self.cache_file = cache_file
        self.secret = secret

    def retrieve_secrets(self):
        if self.cache_file and os.path.isfile(self.cache_file):
            with open(self.cache_file, 'rb') as f:
                if self.secret:
                    raw_data = _decrypt_cipher_text(
                        encrypted_cipher_text=f.read(),
                        secret=self.secret,
                    )
                    return json.loads(raw_data)
                else:
                    return json.load(f)

        request_url = urljoin(self.url, self.concourse_secret_name)
        response = requests.get(request_url, timeout=(4, 31))

        if not response.ok:
        # pylint: enable=no-member
            raise RuntimeError('secrets_server sent {d}: {m}'.format(
                d=response.status_code,
                m=response.content
            ))

        if self.cache_file:
            with open(self.cache_file, 'wb') as f:
                if self.secret:
                    f.write(response.content)
                else:
                    json.dump(response.json(), f)

        if self.secret:
            raw_data = _decrypt_cipher_text(
                encrypted_cipher_text=response.content,
                secret=self.secret,
            )
            return json.loads(raw_data)
        else:
            return response.json()


def _decrypt_cipher_text(encrypted_cipher_text: bytes, secret: model.secret.SecretData):
    if (cipher_alg := secret.cipher_algorithm) is model.secret.Cipher.PLAINTEXT:
        return encrypted_cipher_text
    elif cipher_alg is model.secret.Cipher.AES_ECB:
        pass
    else:
        raise NotImplementedError(cipher_alg)

    from Crypto.Cipher import AES

    cipher = AES.new(key=secret.key, mode=AES.MODE_ECB)

    decrypted_cipher = cipher.decrypt(encrypted_cipher_text)
    try:
        unpadded_cipher = Crypto.Util.Padding.unpad(
            padded_data=decrypted_cipher,
            block_size=AES.block_size,
        )
    except ValueError as ve:
        raise ValueError("Unable to decrypt secret. Key doesn't fit cipher text.") from ve

    return unpadded_cipher


@ensure.ensure_annotations
def encrypt_data(
    key: bytes,
    cipher_algorithm: str | model.secret.Cipher,
    serialized_secret_data: bytes,
) -> bytes:
    cipher_algorithm = model.secret.Cipher(cipher_algorithm)

    if cipher_algorithm is model.secret.Cipher.PLAINTEXT:
        return serialized_secret_data
    elif cipher_algorithm is model.secret.Cipher.AES_ECB:
        secret_key = base64.b64decode(key)
        cipher = AES.new(secret_key, AES.MODE_ECB)
        cipher_text = cipher.encrypt(
            Crypto.Util.Padding.pad(
                data_to_pad=serialized_secret_data,
                block_size=AES.block_size,
            )
        )
    else:
        logger.error(f'cipher algorithm defined is not supported. {cipher_algorithm}')
        raise NotImplementedError(cipher_algorithm)

    return cipher_text



import collections.abc
import dataclasses
import enum
import hashlib
import io
import json
import logging
import tarfile
import zlib

import dacite
import requests

import oci.auth as oa
import oci.client as oc
import oci.convert as oconv
import oci.kaniko as ok
import oci.model as om
import oci.platform as op
import oci.util as ou

logger = logging.getLogger(__name__)

# type-alias for typehints
image_reference = str


class ReplicationMode(enum.Enum):
    '''
    configures the oci artifact replication semantics

    REGISTRY_DEFAULTS:
        do not specify `accept`-header. Depending on OCI-Registry, this will
        in some cases result in "single" oci artifacts (defaulting to variants
        for linux-x86_64) to be replicated, even if multiarch would be available

        for clients that also do not specify the `accept` header, or that require only
        the default image, the result will most closely reflect the one from the source
        registry.

    PREFER_MULTIARCH:
        prefer multiarch if available (by setting `accept` header)

        The replication result will match the source as exact as possible.

    NORMALISE_TO_MULTIARCH:
        like PREFER_MULTIARCH. However, in case only a single artifact is available,
        generate a multiarch manifest with the single artifact as only entry, using the
        `os` and `architecture` attributes from the artifact's cfg-blob.
    '''
    REGISTRY_DEFAULTS = 'registry_defaults'
    PREFER_MULTIARCH = 'prefer_multiarch'
    NORMALISE_TO_MULTIARCH = 'normalise_to_multiarch'

    def accept_header(self):
        if self is ReplicationMode.REGISTRY_DEFAULTS:
            return None
        elif self is ReplicationMode.PREFER_MULTIARCH:
            return om.MimeTypes.prefer_multiarch
        elif self is ReplicationMode.NORMALISE_TO_MULTIARCH:
            return om.MimeTypes.prefer_multiarch
        else:
            raise NotImplementedError(self)


def replicate_artifact(
    src_image_reference: str | om.OciImageReference,
    tgt_image_reference: str | om.OciImageReference,
    credentials_lookup: oa.credentials_lookup=None,
    routes: oc.OciRoutes=oc.OciRoutes(),
    oci_client: oc.Client=None,
    mode: ReplicationMode=ReplicationMode.REGISTRY_DEFAULTS,
    platform_filter: collections.abc.Callable[[om.OciPlatform], bool]=None,
    annotations: dict[str, str]=None,
) -> tuple[requests.Response, str, bytes]:
    '''
    replicate the given OCI Artifact from src_image_reference to tgt_image_reference.

    try to be verbatim, if possible (i.e. target should reflect source as close as
    possible). Whether or not a fully identical replication is possible depends on
    the source artifact and chosen replication `mode`:

    If source artifact is a "legacy / v1" "docker image" (as it used to be
    created from older versions of docker) verbatim replication is not
    possible, because modern (v2) OCI Registries (such as GCR) will not accept
    those manifests. Therefore, conversion to "v2" is required (done
    transparently by this function).

    If source artifact is a "multiarch" image (oci.model.OciImageManifestList), OCI
    registries show different behaviour if ReplicationMode.REGISTRY_DEFAULTS is used.
    Some registries will in this case return a single-image manifest, instead of the
    multiarch-manifest (in this case, the replication result will only be a single-image).

    Use ReplicationMode.PREFER_MULTIARCH or ReplicationMode.NORMALISE_TO_MULTIARCH to
    prevent this.

    If platform_filter is specified (only applied for multi-arch images), the replication
    result will obviously also deviate from src, depending on the filter semantics.

    If `annotations` are passed (which must be a dict w/ plain str->str entries as
    specified by https://github.com/opencontainers/image-spec/blob/main/annotations.md),
    they are updated into the target-manifest. If they are already present, they will be
    overwritten. If existing values are identical, it is tried to avoid to create a "pseudo-diff"
    (i.e. in case the existing values are equal, the manifest will be left untouched).

    pass either `credentials_lookup`, `routes`, OR `oci_client`
    '''
    if not (bool(credentials_lookup) ^ bool(oci_client)):
        raise ValueError('either credentials-lookup + routes, xor client must be passed')

    src_image_reference = om.OciImageReference.to_image_ref(src_image_reference)
    tgt_image_reference = om.OciImageReference.to_image_ref(tgt_image_reference)

    if not oci_client:
        client = oc.Client(
            credentials_lookup=credentials_lookup,
            routes=routes,
        )
    else:
        client = oci_client

    accept = mode.accept_header()

    # we need the unaltered - manifest for verbatim replication
    raw_manifest = client.manifest_raw(
        image_reference=src_image_reference,
        accept=accept,
    ).text
    manifest = json.loads(raw_manifest)
    schema_version = int(manifest['schemaVersion'])
    need_to_synthesise_cfg_blob = False

    if schema_version == 1:
        need_to_synthesise_cfg_blob = True
        manifest = client.manifest(image_reference=src_image_reference)

        logger.warning(
          f'''
          manifest {src_image_reference=} is in legacy-format
          (schemaVersion==1). Cannot verbatimly replicate
          '''
        )

        manifest, _ = oconv.v1_manifest_to_v2(
            manifest=manifest,
            oci_client=client,
            tgt_image_ref=str(tgt_image_reference),
        )

        # we must determine the uncompressed layer-digests to synthesise a valid
        # cfg-blob docker will accept (this means in particular we must download
        # all layers, even if we do not need to upload them)
        need_uncompressed_layer_digests = True
        uncompressed_layer_digests = []
    elif schema_version == 2:
        manifest = json.loads(raw_manifest)
        media_type = manifest.get('mediaType', om.DOCKER_MANIFEST_SCHEMA_V2_MIME)

        if media_type in (
            om.DOCKER_MANIFEST_LIST_MIME,
            om.OCI_IMAGE_INDEX_MIME,
        ):
            # multi-arch
            manifest = dacite.from_dict(
                data_class=om.OciImageManifestList,
                data=manifest,
            )
            manifest: om.OciImageManifestList

            src_ref = om.OciImageReference(image_reference=src_image_reference)
            src_name = src_ref.ref_without_tag
            tgt_ref = om.OciImageReference(image_reference=tgt_image_reference)
            tgt_name = tgt_ref.ref_without_tag

            # try to avoid modifications (from x-serialisation) - unless we have to
            manifest_dirty = False

            # cp manifests to tuple, because we _might_ modify if there is a platform_filter
            for idx, sub_manifest in enumerate(tuple(manifest.manifests)):
                src_reference = f'{src_name}@{sub_manifest.digest}'
                tgt_reference = f'{tgt_name}'

                if platform_filter:
                    platform = op.from_single_image(
                        image_reference=src_reference,
                        oci_client=oci_client,
                        base_platform=sub_manifest.platform,
                    )
                    if not platform_filter(platform):
                        logger.info(f'skipping {platform=} for {src_image_reference=}')
                        manifest_dirty = True
                        manifest.manifests.remove(sub_manifest)
                        continue

                logger.info(f'replicating to {tgt_reference=}')

                res, ref, submanifest_bytes = replicate_artifact(
                    src_image_reference=src_reference,
                    tgt_image_reference=tgt_reference,
                    oci_client=client,
                    annotations=annotations,
                )

                submanifest_digest = f'sha256:{hashlib.sha256(submanifest_bytes).hexdigest()}'
                if submanifest_digest != sub_manifest.digest:
                    patched_sub_manifest = dataclasses.replace(
                        sub_manifest,
                        digest=submanifest_digest,
                        size=len(submanifest_bytes),
                    )
                    manifest.manifests.remove(sub_manifest)
                    manifest.manifests.insert(idx, patched_sub_manifest)
                    manifest_dirty = True

            if annotations:
                # try to avoid unnecessary changes by x-serialisation - only add values if
                # they are either new or different
                for k, v in annotations.items():
                    if manifest.annotations.get(k) == v:
                        continue
                    else:
                        manifest.annotations[k] = v
                        manifest_dirty = True

            if manifest_dirty:
                raw_manifest = json.dumps(manifest.as_dict())

            if not tgt_image_reference.has_tag:
                digest = f'sha256:{hashlib.sha256(raw_manifest.encode("utf-8")).hexdigest()}'
                tgt_image_reference = om.OciImageReference(
                    f'{tgt_image_reference.ref_without_tag}@{digest}',
                )

            res = client.put_manifest(
                image_reference=tgt_image_reference,
                manifest=raw_manifest,
            )

            return res, tgt_image_reference, raw_manifest.encode('utf-8')

        elif media_type in (
            om.OCI_MANIFEST_SCHEMA_V2_MIME,
            om.DOCKER_MANIFEST_SCHEMA_V2_MIME,
        ):
            if mode is ReplicationMode.NORMALISE_TO_MULTIARCH:
                if not src_image_reference.has_digest_tag:
                    src_image_reference = om.OciImageReference.to_image_ref(
                        oci_client.to_digest_hash(
                            image_reference=src_image_reference,
                        )
                    )
                platform = op.from_single_image(
                    image_reference=src_image_reference,
                    oci_client=oci_client,
                )
                # force usage of digest-tag (symbolic tag required for manifest-list
                tgt_image_ref = \
                    f'{tgt_image_reference.ref_without_tag}@{src_image_reference.tag}'

                res, ref, manifest_bytes = replicate_artifact(
                    src_image_reference=src_image_reference,
                    tgt_image_reference=tgt_image_ref,
                    oci_client=oci_client,
                    annotations=annotations,
                )

                manifest_list = om.OciImageManifestList(
                    manifests=[
                        om.OciImageManifestListEntry(
                            digest=f'sha256:{hashlib.sha256(manifest_bytes).hexdigest()}',
                            mediaType=media_type,
                            size=len(manifest_bytes),
                            platform=platform,
                        ),
                    ],
                    mediaType=om.DOCKER_MANIFEST_LIST_MIME,
                )

                manifest_list_bytes = json.dumps(
                    manifest_list.as_dict(),
                ).encode('utf-8')

                res = oci_client.put_manifest(
                    image_reference=tgt_image_reference,
                    manifest=manifest_list_bytes,
                )

                return res, tgt_image_reference, manifest_list_bytes

            manifest = dacite.from_dict(
                data_class=om.OciImageManifest,
                data=json.loads(raw_manifest)
            )
            need_uncompressed_layer_digests = False
            uncompressed_layer_digests = None
        else:
            raise NotImplementedError(f'{media_type=}')
    else:
      raise NotImplementedError(schema_version)

    for idx, layer in enumerate(manifest.blobs()):
        # need to specially handle cfg-blob (may be absent for v2 / legacy images)

        is_cfg_blob = idx == 0
        if is_cfg_blob and need_to_synthesise_cfg_blob:
            # if we need(ed) to synthesise cfg-blob (because source-image contained a v1-manifest)
            # then there will never be a cfg-blob in src.
            # -> silently skip to avoid emitting a confusing, but unhelpful warning
            logger.debug(f'{src_image_reference=} - synthesised cfg-blob - skipping replication')
            continue

        head_res = client.head_blob(
            image_reference=tgt_image_reference,
            digest=layer.digest,
        )
        if head_res.ok:
            if not need_uncompressed_layer_digests:
                logger.info(f'skipping blob download {layer.digest=} - already exists in tgt')
                continue # no need to download if blob already exists in tgt
            elif not is_cfg_blob:
                # we will not need to re-upload, however we do need the uncompressed digest
                blob_res = client.blob(
                    image_reference=src_image_reference,
                    digest=layer.digest,
                    absent_ok=is_cfg_blob,
                )

                layer_hash = hashlib.sha256()
                decompressor = zlib.decompressobj(wbits=zlib.MAX_WBITS | 16)

                for chunk in blob_res.iter_content(chunk_size=4096):
                    layer_hash.update(decompressor.decompress(chunk))

                uncompressed_layer_digests.append(f'sha256:{layer_hash.hexdigest()}')
                continue # we may still skip the upload, of course

        # todo: consider silencing warning if we do v1->v2-conversion (cfg-blob will never exist
        #       in this case
        blob_res = client.blob(
            image_reference=src_image_reference,
            digest=layer.digest,
            absent_ok=is_cfg_blob,
        )
        if not blob_res and is_cfg_blob:
            # fallback to non-verbatim replication; synthesise cfg
            logger.warning(
                'falling back to non-verbatim replication '
                f'{src_image_reference=} {tgt_image_reference=}'
            )
            need_to_synthesise_cfg_blob = True
            continue

        if need_uncompressed_layer_digests:
            uncompressed_layer_hash = hashlib.sha256()
            decompressor = zlib.decompressobj(wbits=zlib.MAX_WBITS | 16)

            def intercept_chunks(blob_res):
                for chunk in blob_res.iter_content(chunk_size=4096):
                    uncompressed_layer_hash.update(decompressor.decompress(chunk))
                    yield chunk

                uncompressed_layer_digests.append(f'sha256:{uncompressed_layer_hash.hexdigest()}')

            blob_res = intercept_chunks(blob_res=blob_res)

        client.put_blob(
            image_reference=tgt_image_reference,
            digest=layer.digest,
            octets_count=layer.size,
            data=blob_res,
        )

    if need_to_synthesise_cfg_blob:
        fake_cfg_dict = json.loads(json.loads(raw_manifest)['history'][0]['v1Compatibility'])

        # patch-in uncompressed layer-digests
        fake_cfg_dict['rootfs'] = {
            'diff_ids': uncompressed_layer_digests,
            'type': 'layers',
        }

        fake_cfg_raw = json.dumps(fake_cfg_dict).encode('utf-8')

        client.put_blob(
            image_reference=tgt_image_reference,
            digest=(cfg_digest := f'sha256:{hashlib.sha256(fake_cfg_raw).hexdigest()}'),
            octets_count=len(fake_cfg_raw),
            data=fake_cfg_raw,
        )

        manifest_dict = manifest.as_dict()
        # patch-on altered cfg-digest
        manifest_dict['config']['digest'] = cfg_digest
        manifest_dict['config']['size'] = len(fake_cfg_raw)
        raw_manifest = json.dumps(manifest_dict)

    if annotations:
        manifest_dirty = False

        for k, v in annotations.items():
            if manifest.annotations.get(k) == v:
                continue
            else:
                manifest.annotations[k] = v
                manifest_dirty = True

        if manifest_dirty:
            raw_manifest = json.dumps(manifest.as_dict())

    if not tgt_image_reference.has_tag:
        digest = f'sha256:{hashlib.sha256(raw_manifest.encode("utf-8")).hexdigest()}'
        tgt_image_reference = f'{tgt_image_reference.ref_without_tag}@{digest}'

    res = client.put_manifest(
        image_reference=tgt_image_reference,
        manifest=raw_manifest,
    )

    return res, tgt_image_reference, raw_manifest.encode('utf-8')


def replicate_blobs(
    src_ref: str,
    src_oci_manifest: om.OciImageManifest,
    tgt_ref: str,
    oci_client: oc.Client,
    blob_overwrites: dict[om.OciBlobRef, bytes | io.BytesIO],
) -> om.OciImageManifest:
    '''
    replicates blobs from given oci-image-ref to the specified target-ref, optionally replacing
    the specified blobs. This is particularly useful for replacing some "special" blobs, such
    as a component-descriptor layer blob or the config-blob.

    Note that the uploaded artifact must be finalised after the upload by a "manifest-put".
    '''
    blob_overwrites = {k.digest:v for k,v in blob_overwrites.items()}

    def replicate_blob(blob: om.OciBlobRef) -> om.OciBlobRef:
        if blob_overwrite_bytes := blob_overwrites.get(blob.digest):
            logger.info(f'Replicate with overwriting {blob=}')

            if hasattr(blob_overwrite_bytes, 'read'):
                digest = hashlib.sha256()
                blob_overwrite_bytes.seek(0)
                while (chunk := blob_overwrite_bytes.read(4096)):
                    digest.update(chunk)
                digest = f'sha256:{digest.hexdigest()}'
                octets_count = blob_overwrite_bytes.tell()
                blob_overwrite_bytes.seek(0)
            else:
                digest = f'sha256:{hashlib.sha256(blob_overwrite_bytes).hexdigest()}'
                octets_count = len(blob_overwrite_bytes)

            oci_client.put_blob(
                image_reference=tgt_ref,
                digest=digest,
                octets_count=octets_count,
                data=blob_overwrite_bytes,
            )
            return om.OciBlobRef(
                digest=digest,
                mediaType=blob.mediaType, #XXX: pass-in new media type?
                size=octets_count,
            )
        else:
            digest = blob.digest

            src_blob: requests.models.Response = oci_client.blob(
                image_reference=src_ref,
                digest=digest,
            )

            octets_count = int(src_blob.headers['Content-Length'])

            oci_client.put_blob(
                image_reference=tgt_ref,
                digest=digest,
                octets_count=octets_count,
                data=src_blob,
            )
            return om.OciBlobRef(
                digest=digest,
                mediaType=blob.mediaType,
                size=octets_count,
            )

    return om.OciImageManifest(
        config=replicate_blob(src_oci_manifest.config),
        layers=[replicate_blob(blob) for blob in src_oci_manifest.layers],
    )


def publish_container_image_from_kaniko_tarfile(
    image_tarfile_path: str,
    oci_client: oc.Client,
    image_reference: str,
    additional_tags: list[str]=(),
    manifest_mimetype: str=om.OCI_MANIFEST_SCHEMA_V2_MIME,
):
    image_reference = ou.normalise_image_reference(image_reference=image_reference)
    image_name = image_reference.rsplit(':', 1)[0]
    image_references = (image_reference,) + tuple([f'{image_name}:{tag}' for tag in additional_tags])

    with ok.read_kaniko_image_tar(tar_path=image_tarfile_path) as image:
        chunk_size = 1024 * 1024
        for kaniko_blob in image.blobs():
            oci_client.put_blob(
                image_reference=image_reference,
                digest=kaniko_blob.digest_str(),
                octets_count=kaniko_blob.size,
                data=kaniko_blob,
                max_chunk=chunk_size,
            )

        # optionally patch manifest's mimetype (e.g. required for docker-hub)
        manifest_dict = image.oci_manifest().as_dict()
        manifest_dict['mediaType'] = manifest_mimetype

        manifest_bytes = json.dumps(
            manifest_dict,
        ).encode('utf-8')

        for tgt_ref in image_references:
            logger.info(f'publishing manifest {tgt_ref=}')
            oci_client.put_manifest(
                image_reference=tgt_ref,
                manifest=manifest_bytes,
            )


def image_layers_as_tarfile_generator(
    image_reference: str,
    oci_client: oc.Client,
    chunk_size=tarfile.RECORDSIZE,
    include_config_blob=True,
    fallback_to_first_subimage_if_index=False,
) -> collections.abc.Generator[bytes, None, None]:
    '''
    returns a generator yielding a tar-archive with the passed oci-image's layer-blobs as
    members. This is somewhat similar to the result of a `docker save` with the notable difference
    that the cfg-blob is discarded.
    This function is useful to e.g. upload file system contents of an oci-container-image to some
    scanning-tool (provided it supports the extraction of tar-archives)
    If include_config_blob is set to False the config blob will be ignored.

    If fallback_to_first_subimage_if_index is set to True, in case of oci-image-manifest-list the
    first sub-manifest is taken.
    '''
    manifest = oci_client.manifest(
        image_reference=image_reference,
        accept=om.MimeTypes.prefer_multiarch,
    )

    image_reference = om.OciImageReference.to_image_ref(image_reference)

    if fallback_to_first_subimage_if_index and isinstance(manifest, om.OciImageManifestList):
        logger.warn(
            f'image-index handling not fully implemented - will only scan first image, '
            f'{image_reference=}, {manifest.mediaType=}'
        )
        manifest_ref = manifest.manifests[0]
        manifest = oci_client.manifest(
            image_reference=f'{image_reference.ref_without_tag}@{manifest_ref.digest}',
        )

    offset = 0
    for blob in manifest.blobs() if include_config_blob else manifest.layers:
        logger.debug(f'getting blob {blob.digest}')
        if not include_config_blob:
            logger.debug('skipping config blob')
        tarinfo = tarfile.TarInfo(name=blob.digest + '.tar') # note: may be gzipped
        tarinfo.size = blob.size
        tarinfo.offset = offset
        tarinfo.offset_data = offset + tarfile.BLOCKSIZE

        offset += blob.size + tarfile.BLOCKSIZE

        tarinfo_bytes = tarinfo.tobuf()
        yield tarinfo_bytes

        uploaded_bytes = len(tarinfo_bytes)
        for chunk in oci_client.blob(
            image_reference=image_reference,
            digest=blob.digest,
            stream=True,
            ).iter_content(chunk_size=chunk_size):
            uploaded_bytes += len(chunk)
            yield chunk

        # need to pad full blocks w/ NUL-bytes
        if (missing := tarfile.BLOCKSIZE - (uploaded_bytes % tarfile.BLOCKSIZE)):
            offset += missing
            yield tarfile.NUL * missing

    # tarchives should be terminated w/ two empty blocks
    yield tarfile.NUL * tarfile.BLOCKSIZE * 2



'''
utils for processing image-tar as created by kaniko

see:
    https://github.com/GoogleContainerTools/kaniko
'''

import collections.abc
import contextlib
import dataclasses
import io
import json
import tarfile
import threading

import dacite

import oci.model


@dataclasses.dataclass
class KanikoManifest:
  Config: str # <algorithm>:<digest>
  RepoTags: list[str]
  Layers: list[str]


@dataclasses.dataclass
class _KanikoBlob(io.BytesIO):
  read_chunk: collections.abc.Callable[[int, int], bytes] # see _KanikoImageReadCtx._read_chunk
  offset: int
  name: str
  size: int
  hash_algorithm: str = 'sha256'
  seek_pos = 0

  def __post_init__(self):
    self.__iter__ = self.iter_contents

  def __len__(self):
    return self.size

  def read(self, size: int=-1):
    if size == -1:
      offset = self.seek_pos + self.offset
      self.seek_pos = size - 1

      return self.read_chunk(
        offset=offset,
        length=self.size,
      )

    if self.seek_pos + 1 == self.size:
        return b''

    offset = self.offset + self.seek_pos
    length = min(size, self.size - self.seek_pos)
    self.seek_pos += size
    if self.seek_pos + 1 > self.size:
        self.seek_pos = self.size - 1

    return self.read_chunk(
        offset=offset,
        length=length,
    )

  def tell(self):
    return self.seek_pos

  def iter_contents(self, chunk_size=1024 * 1024):
    remaining = self.size
    read = 0

    while remaining > 0:
      this_chunk = min(chunk_size, remaining)

      yield self.read_chunk(
        offset=self.offset + read,
        length=this_chunk
      )

      remaining -= this_chunk
      read += this_chunk

  def digest_hash(self):
    if ':' in self.name:
      return self.name.split(':')[-1]
    elif '.' in self.name:
      return self.name.split('.')[0]

  def digest_str(self):
    return f'{self.hash_algorithm}:{self.digest_hash()}'.strip()


class _KanikoImageReadCtx:
  def __init__(
      self,
      img_tarfile: tarfile.TarFile,
  ):
    self.tarfile = img_tarfile
    self.fileobj = img_tarfile.fileobj
    self.kaniko_manifest = self._kaniko_manifest()
    self._lock = threading.Lock()

  def _kaniko_manifest(self):
    manifest_info = self.tarfile.getmember('manifest.json')
    self.fileobj.seek(manifest_info.offset_data)
    manifest_raw = self.fileobj.read(manifest_info.size)
    manifest_list = json.loads(manifest_raw.decode('utf-8'))

    if not (leng := len(manifest_list)) == 1:
      raise NotImplementedError(leng)

    manifest = manifest_list[0]

    return dacite.from_dict(
        data_class=KanikoManifest,
        data=manifest,
    )

  def _read_chunk(self, offset: int, length: int):
    with self._lock:
      self.fileobj.seek(offset)
      return self.fileobj.read(length)

  def cfg_blob(self):
    cfg_info = self.tarfile.getmember(name=self.kaniko_manifest.Config)

    # HACK: unfortunately, we need to patch the cfg-blob in case there are
    # less history-entries than layer-blobs
    def _mk_wrapper():
        return  _KanikoBlob(
          read_chunk=self._read_chunk,
          offset=cfg_info.offset_data,
          name=cfg_info.name,
          size=cfg_info.size,
          hash_algorithm=cfg_info.name.split(':')[0],
        )

    cfg_blob_bytes = _mk_wrapper().read()
    cfg_blob_dict = json.loads(cfg_blob_bytes)

    layers_count = len(tuple(self.layer_blobs()))

    if layers_count <= (history_leng := len((history := cfg_blob_dict['history']))):
        # no patching needed
        print('no patching required')
        return _mk_wrapper()

    missing_history_count = layers_count - history_leng
    last_history_entry = history[-1] # hackily cp last entry
    for _ in range(missing_history_count):
        history.append(last_history_entry)

    cfg_blob_dict['history'] = history
    cfg_blob_bytes = json.dumps(cfg_blob_dict).encode('utf-8')

    import hashlib
    cfg_hash = hashlib.sha256(cfg_blob_bytes).hexdigest()

    cfg_blob_buf = io.BytesIO(cfg_blob_bytes)
    cfg_blob_buf.seek(0)

    # now we need to fake a kaniko-blob-wrapper
    cfg_blob_wrapper =  _KanikoBlob(
      read_chunk=self._read_chunk,
      offset=cfg_info.offset_data,
      name=f'sha256:{cfg_hash}', # XXX name is parsed again for digest_str()
      size=len(cfg_blob_bytes),
      hash_algorithm=cfg_info.name.split(':')[0],
    )
    cfg_blob_wrapper.read = cfg_blob_buf.read

    return cfg_blob_wrapper

  def layer_blobs(self):
    for layer_name in self.kaniko_manifest.Layers:
      layer_info = self.tarfile.getmember(name=layer_name)

      yield _KanikoBlob(
        read_chunk=self._read_chunk,
        offset=layer_info.offset_data,
        name=layer_name,
        size=layer_info.size,
        hash_algorithm='sha256', # XXX hardcode for now
      )

  def blobs(self):
    yield self.cfg_blob()
    yield from self.layer_blobs()

  def oci_manifest(self):
    cfg = self.cfg_blob()

    return oci.model.OciImageManifest(
      mediaType=oci.model.DOCKER_MANIFEST_SCHEMA_V2_MIME,
      config=oci.model.OciBlobRef(
        digest=cfg.digest_str(),
        mediaType='application/vnd.docker.container.image.v1+json',
        size=cfg.size,
      ),
      layers=[
        oci.model.OciBlobRef(
          digest=layer.digest_str(),
          # hard-code kaniko's output mimetype for now
          mediaType='application/vnd.docker.image.rootfs.diff.tar.gzip',
          size=layer.size,
        ) for layer in self.layer_blobs()
      ],
    )


@contextlib.contextmanager
def read_kaniko_image_tar(tar_path: str):
  '''
  @param tar_path: path to image-tar created by kaniko
  '''
  with tarfile.open(name=tar_path, mode='r:*') as tf:
    yield _KanikoImageReadCtx(img_tarfile=tf)



import dataclasses
import datetime
import logging
import typing

import cfg_mgmt.model as cmm


logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class CcCfgComplianceStorageResponsibles:
    creation_date: str
    compliant: int
    noncompliant: int
    responsible_name: str
    responsible_type: str
    url: str

    @staticmethod
    def create(
        url: str,
        responsible: cmm.CfgResponsible,
        compliant_count: int,
        non_compliant_count: int,
    ):
        '''
        convenience method to create a `CcCfgComplianceStorageResponsibles`
        '''
        return CcCfgComplianceStorageResponsibles(
            creation_date=datetime.datetime.now().isoformat(),
            compliant=compliant_count,
            noncompliant=non_compliant_count,
            responsible_name=responsible.name,
            responsible_type=responsible.type.value,
            url=url,
        )


@dataclasses.dataclass(frozen=True)
class CcCfgComplianceStatus:
    '''
    represents counts for (non)compliant cfg_elements for a url (element_storage)
    '''
    url: str
    compliant: int
    noncompliant: int
    creation_date: str

    @staticmethod
    def create(
        url: str,
        compliant_count: int,
        non_compliant_count: int,
    ):
        '''
        convenience method to create a `CcCfgComplianceStatus`
        '''
        return CcCfgComplianceStatus(
            url=url,
            compliant=compliant_count,
            noncompliant=non_compliant_count,
            creation_date=datetime.datetime.now().isoformat(),
        )


@dataclasses.dataclass(frozen=True)
class CcCfgComplianceResponsible:
    '''
    represents compliance status for a cfg_element per responsible
    '''
    creation_date: str
    element_name: str
    element_type: str
    element_storage: str
    is_compliant: bool
    responsible_name: typing.List[str]
    responsible_type: typing.List[str]
    rotation_method: str
    non_compliant_reasons: typing.List[str]

    @staticmethod
    def create(
        element_name: str,
        element_type: str,
        element_storage: str,
        is_compliant: bool,
        responsible: cmm.CfgResponsibleMapping,
        rotation_method: cmm.RotationMethod,
        non_compliant_reasons: typing.List[cmm.CfgElementPolicyViolation]
    ):
        '''
        convenience method to create a `CcCfgComplianceResponsible`
        '''
        names = []
        types = []
        if responsible:
            names = [resp.name for resp in responsible.responsibles]
            types = [resp.type.value for resp in responsible.responsibles]

        reasons = [
            reason.value
            for reason in non_compliant_reasons
        ]
        return CcCfgComplianceResponsible(
            creation_date=datetime.datetime.now().isoformat(),
            element_name=element_name,
            element_type=element_type,
            element_storage=element_storage,
            is_compliant=is_compliant,
            responsible_name=names,
            responsible_type=types,
            rotation_method=rotation_method.value,
            non_compliant_reasons=reasons,
        )


def index_name(
    obj: typing.Union[CcCfgComplianceStatus, CcCfgComplianceResponsible],
) -> str:
    if isinstance(obj, CcCfgComplianceResponsible):
        return 'cc_cfg_compliance_responsible'

    if isinstance(obj, CcCfgComplianceStatus):
        return 'cc_cfg_compliance_status'

    if isinstance(obj, CcCfgComplianceStorageResponsibles):
        return 'cc_cfg_compliance_storage_responsibles'

    raise NotImplementedError(obj)
